# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Cylc provides support for the following job runner handlers.

.. autosummary::
   :toctree: job-runner-handlers
   :template: automodule_job_runner_handlers.rst

   cylc.flow.job_runner_handlers.at
   cylc.flow.job_runner_handlers.background
   cylc.flow.job_runner_handlers.loadleveler
   cylc.flow.job_runner_handlers.lsf
   cylc.flow.job_runner_handlers.moab
   cylc.flow.job_runner_handlers.pbs
   cylc.flow.job_runner_handlers.sge
   cylc.flow.job_runner_handlers.slurm
   cylc.flow.job_runner_handlers.slurm_packjob

"""
